#############
# $Id: harold_calibration_bam.pm 14173 2011-09-13 09:45:29Z ajb $
# Created By: gq1
# Last Maintained By: $Author: ajb $
# Created On: 2011-07-27
# Last Changed On: $Date: 2011-09-13 10:45:29 +0100 (Tue, 13 Sep 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-30.0/lib/npg_pipeline/analysis/harold_calibration_bam.pm $

package npg_pipeline::analysis::harold_calibration_bam;
use Moose;
use Carp;
use English qw{-no_match_vars};
use Cwd;      # for getcwd
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 14173 $ =~ /(\d+)/mxs; $r; };
use npg_common::VersionComparison;

Readonly::Scalar our $PB_ALIGN_BAM_PREFIX => q{pb_align_};

Readonly::Scalar our $BAM_MERGER_CMD => q{/software/bin/java -jar /software/solexa/bin/aligners/illumina2bam/current/BamMerger.jar CREATE_MD5_FILE=true VALIDATION_STRINGENCY=SILENT KEEP=true I=/dev/stdin};

extends qw{npg_pipeline::analysis::bustard};

with qw{
  npg_pipeline::roles::business::harold_calibration_reqs
  npg_common::roles::run::intensities::config
  npg_pipeline::roles::business::pre_exec_strings
};

=head1 NAME

npg_pipeline::analysis::harold_calibration_bam

=head1 VERSION

$LastChangedRevision: 14173 $

=head1 SYNOPSIS

  my $oHaroldCalibration = npg_pipeline:analysis::harold_calibration_bam->new();

=head1 DESCRIPTION

Object runner to launch internal calibration instead of CASAVA based calibration

=head1 SUBROUTINES/METHODS

=head2 generate_alignment_files

submit the jobs which will generate bam alignment files ready to pass onto calibration table generator

  my $aJobIds = $oHaroldCalibration->generate_alignment_files({
    required_job_completion => $sJobRequirenmentString,
  });

=cut

sub generate_alignment_files {
  my ( $self, $arg_refs ) = @_;
  my $job_ids = [];

  if ( $self->no_eland() || $self->no_recalibration() ) {
    $self->log( q{This has been set to run with no recalibration step} );
    return ();
  }

  my @positions = $self->positions( $arg_refs );

  my %positions = map {$_=>1} @positions;

  my $job_dependencies = $arg_refs->{required_job_completion};

  my $run = $self->run();

  my $cur_dir = getcwd();            # save for later
  my $dir = $self->basecall_path();
  $self->make_log_dir( $dir ); # create a log directory within BaseCalls

  # create the calibration directory
  my $pb_cal_dir = $self->create_pb_calibration_directory();

  $self->_set_recalibrated_path( $self->pb_cal_path() );

  $self->log( qq{Changing to $pb_cal_dir} );
  chdir $pb_cal_dir or croak "could not cd to $pb_cal_dir"; # change to the calibration directory

  foreach my $run_lane ( @{$run->run_lanes()} ) {

    if( !$positions{$run_lane->position()} ){
       next;
    }

    if ( $self->no_recalibration() ) {
      $self->log( q{no_recalibration option selected, no processing of calibration tables for any lanes performed} );
      last;
    }

    $self->_generate_alignment_file_per_lane( {
      run_lane => $run_lane,
      job_ids => $job_ids,
      job_dependencies => $job_dependencies,
    } );
  }

  $self->log( qq{Changing back to $cur_dir} );
  chdir $cur_dir or croak "could not cd back to $cur_dir";  # go back home

  return @{ $job_ids };
}

=head2 generate_calibration_table

submit the bsub jobs which will create the calibration tables, returning an array of job_ids.

  my $aJobIds = $oHaroldCalibration->generate_calibration_table({
    required_job_completion => $sJobRequirenmentString,
  });

=cut

sub generate_calibration_table {
  my ($self, $arg_refs) = @_;

  if ( $self->no_eland() || $self->no_recalibration() ) {
    $self->log( q{This has been set to run with no recalibration step} );
    return ();
  }

  my @positions = $self->positions( $arg_refs );

  my %positions = map {$_=>1} @positions;

  my $job_dependencies = $arg_refs->{required_job_completion};

  my $run     = $self->run();
  my $job_ids = [];

  my $id_run = $self->id_run();

  my $cur_dir = getcwd();            # save for later
  my $dir = $self->basecall_path();

  # create the calibration directory
  my $pb_cal_dir = $self->create_pb_calibration_directory();

  $self->_set_recalibrated_path( $self->pb_cal_path() );

  $self->log( qq{Changing to $pb_cal_dir} );
  chdir $pb_cal_dir or croak "could not cd to $pb_cal_dir"; # change to the calibration directory

  foreach my $run_lane ( @{$run->run_lanes()} ) {

    if( !$positions{$run_lane->position()} ){
       next;
    }

    if ( $self->no_recalibration() ) {
      $self->log( q{no_recalibration option selected, no processing of calibration tables for any lanes performed} );
      last;
    }

    $self->_generate_calibration_table_per_lane( {
      run_lane => $run_lane,
      job_ids => $job_ids,
      job_dependencies => $job_dependencies,
    } );
  }

  $self->log( qq{Changing back to $cur_dir} );
  chdir $cur_dir or croak "could not cd back to $cur_dir";  # go back home

  return @{ $job_ids };
}

=head2 generate_recalibrated_qseqs

submit the bsub jobs which will recalibrate the lanes, returning an array of job_ids.

  my $aJobIds = $oHaroldCalibration->generate_recalibrated_qseqs({
    required_job_completion => $sJobRequirenmentString,
  });

=cut

sub generate_recalibrated_qseqs {
  my ($self, $arg_refs) = @_;

  if ( $self->no_eland() || $self->no_recalibration() ) {
    $self->log( q{This has been set to run with no recalibration step} );
    return ();
  }

  my @positions = $self->positions( $arg_refs );

  my %positions = map {$_=>1} @positions;

  my $pb_cal_dir = $self->pb_cal_path();

  if ( ! $self->directory_exists($pb_cal_dir) ) {
    $self->log( qq{$pb_cal_dir does not exist, not executing jobs} );
    return ();
  }

  my $id_run = $self->id_run();

  my ( $read_1_calibration_table, $read_2_calibration_table );

  # if there is a control lane, then this will be used for any lanes with no reference sequence
  $read_1_calibration_table = $self->calibration_table_name( {
      id_run => $id_run,
      read => 1,
  } );

  $read_2_calibration_table = $self->calibration_table_name( {
    id_run => $id_run,
    read => 2,
  } );


  my $job_dependencies = $arg_refs->{required_job_completion};

  my $run     = $self->run();
  my $job_ids = [];

  my $cur_dir = getcwd();            # save for later

  $self->log( qq{Changing to $pb_cal_dir} );
  chdir $pb_cal_dir or croak "could not cd to $pb_cal_dir"; # change to the calibration directory

  foreach my $run_lane ( @{$run->run_lanes()} ) {

    if( !$positions{$run_lane->position()} ){
       next;
    }

    my $arg_ref_hash = {
      read_1_calibration_table => $read_1_calibration_table,
      read_2_calibration_table => $read_2_calibration_table,
      job_ids => $job_ids,
      run_lane => $run_lane,
      job_dependencies => $job_dependencies,
    };
    $self->_generate_recalibrated_qseqs_per_lane( $arg_ref_hash );
  }

  $self->log( qq{Changing back to $cur_dir} );
  chdir $cur_dir or croak "could not cd back to $cur_dir";  # go back home

  return @{ $job_ids };
}

##########
# private methods

sub _generate_recalibrated_qseqs_per_lane {
  my ( $self, $arg_refs ) = @_;

  my $id_run = $self->id_run();

  my $run_lane = $arg_refs->{run_lane};
  my $lane = $run_lane->position();

  my $job_ids = $arg_refs->{job_ids};

  my $cal_table_1_to_use = $arg_refs->{read_1_calibration_table};
  my $cal_table_2_to_use = $arg_refs->{read_2_calibration_table};

  my $ref_seq = $self->ref_seq( {
      lane => {
        id_run => $id_run,
        position => $lane,
      }
    } );

  # if there is a reference sequence, or the lane has spiked in phix, then use lanes own caltable
  if ( $ref_seq || $run_lane->is_spiked_phix() ) {
    $cal_table_1_to_use = $self->calibration_table_name( {
      id_run => $id_run,
      read => 1,
      position => $lane,
    } );
    if ( $self->is_paired_read() ) {
      $cal_table_2_to_use = $self->calibration_table_name( {
         id_run => $id_run,
         read => 2,
         position => $lane,
      } );
    }
  }

  my $mode;
  if ( $self->has_pb_cal_mode() ) {
    $mode = $self->pb_cal_mode();
  } elsif ( $cal_table_1_to_use eq $arg_refs->{read_1_calibration_table} ) {
    # using control lane recal table
    if ( $cal_table_1_to_use !~ /cycle/xms ) {
      $mode = 2;
    }
  } else {};

  my $clmode;
  if ( $arg_refs->{read_1_calibration_table} !~ /cycle/xms ) {
    $clmode = 2;
  }

  my $args_bam;

  if( !$self->is_paired_read() ){

     $args_bam= {
       position => $lane,
       job_dependencies => $arg_refs->{job_dependencies},
       ct => $cal_table_1_to_use,
       clct => $arg_refs->{read_1_calibration_table},
       mode => $mode,
       clmode => $clmode,
     };
  }else {

     $args_bam= {
       position => $lane,
       job_dependencies => $arg_refs->{job_dependencies},
       ct1 => $cal_table_1_to_use,
       clct1 => $arg_refs->{read_1_calibration_table},
       ct2 => $cal_table_2_to_use,
       clct2 => $arg_refs->{read_2_calibration_table},
       mode => $mode,
       clmode => $clmode,
     };
  }


  # only submit if there is a recalibration to do, either against a control lane or a reference
  if ( $ref_seq || $self->pick_control_lane() ) {

    my $bsub_command = $self->_recalibration_bsub_command( $args_bam );
    if ( $self->verbose() ) {
      $self->log( $bsub_command );
    }

    push @{ $job_ids }, $self->submit_bsub_command( $bsub_command );

  }

  return;
}

sub _generate_calibration_table_per_lane {
  my ( $self, $arg_refs ) = @_;

  my $run_lane = $arg_refs->{run_lane};
  my $job_ids  = $arg_refs->{job_ids};
  my $id_run   = $self->id_run();

  my $library  = $run_lane->library();  # only one of $library, $control and $pool will be defined
  my $control  = $run_lane->control();
  my $pool     = $run_lane->pool();

  my $entity   = $run_lane->entity();   # entity is library || control || pool
  my $position = $run_lane->position();
  my $force_control = $self->has_control_lane_position() ? $self->control_lane_position() : 0;

  my $ref_seq;

  my $is_control = 0;
  my $is_spiked_phix = 0;
  # if the lane is a control, or has been declared the forced control
  if ( ! $self->no_control_lane() && ( $control || ( $force_control == $position ) )  ) {

    $ref_seq  = $self->ref_seq( {
      control => 1,
    } );
    $self->log( q{control - recalibrate control lane } . $position );
    $is_control = 1;

  # else see if the lane has a reference sequence
  } else {

    if ( $run_lane->is_spiked_phix() ) {
      $self->log( q{noncontrol lane } . $position . q{ contains spiked-in phix} );
      $ref_seq = $self->ref_seq( {
        is_spiked_phix => 1,
      } );
      $is_spiked_phix++;
    } else {

      my $log_msg = q{control - noncontrol lane} . $position . qq{\n\tchecking to see if reference};
      $self->log( $log_msg );

      $ref_seq = $self->ref_seq( {
        lane => {
          id_run => $id_run,
          position => $position,
        }
      } ) || q{};

    }

    if ( $ref_seq ) {
      $self->log( qq{$id_run : $position : $ref_seq} );
    } else {
      $self->log( q{no ref seq found, no recalibration table generated for lane } . $position );
      return;
    }

  }

  if ( ! $ref_seq ) {
    return;
  }

  my $bsub_command = $self->_calibration_table_bsub_command( {
    position => $position,
    read => 1,
    job_dependencies => $arg_refs->{job_dependencies},
    control => $is_control,
    is_spiked_phix => $is_spiked_phix,
  } );

  if ( $self->verbose() ) {
    $self->log( $bsub_command );
  }

  push @{ $job_ids }, $self->submit_bsub_command($bsub_command);

  if ( $self->is_paired_read() ) {
    my $args = {
      position => $position,
      read => 2,
      job_dependencies => $arg_refs->{job_dependencies},
      control => $is_control,
      is_spiked_phix => $is_spiked_phix,
    };

#    if ( $self->is_indexed() ) {
#      $args->{read} = $self->general_values_conf()->{illumina_read_two};
#    }

    $bsub_command = $self->_calibration_table_bsub_command( $args );

    if ( $self->verbose() ) {
      $self->log( $bsub_command );
    }

    push @{ $job_ids }, $self->submit_bsub_command($bsub_command);
  }

  return;
}

# generate the alignment file
sub _generate_alignment_file_per_lane {
  my ( $self, $arg_refs ) = @_;

  my $run_lane = $arg_refs->{run_lane};
  my $job_ids  = $arg_refs->{job_ids};
  my $id_run   = $self->id_run();

  my $library  = $run_lane->library();  # only one of $library, $control and $pool will be defined
  my $control  = $run_lane->control();
  my $pool     = $run_lane->pool();

  my $entity   = $run_lane->entity();   # entity is library || control || pool
  my $position = $run_lane->position();
  my $force_control = $self->has_control_lane_position() ? $self->control_lane_position() : 0;

  my $ref_seq;

  my $is_control = 0;
  my $is_spiked_phix = 0;

  # if the lane is a control, or has been declared the forced control
  if ( ! $self->no_control_lane() && ( $control || ( $force_control == $position ) )  ) {

    $ref_seq  = $self->ref_seq( {
      control => 1,
    } );
    $self->log( q{control - alignment file done with control reference } . $position );
    $is_control = 1;

  # else see if the lane has a reference sequence
  } else {

    if ( $run_lane->is_spiked_phix() ) {
      $self->log( q{noncpntrol lane } . $position . q{ contains spiked-in phix} );
      $ref_seq = $self->ref_seq( {
        is_spiked_phix => 1,
      } );
      $is_spiked_phix++;
    } else {

      my $log_msg = q{noncontrol lane} . $position . qq{\n\tchecking to see if reference};
      $self->log( $log_msg );

      $ref_seq = $self->ref_seq( {
        lane => {
          id_run => $id_run,
          position => $position,
        }
      } ) || q{};

    }

    if ( $ref_seq ) {
      $self->log( qq{$id_run : $position : $ref_seq} );
    } else {
      $self->log( q{no ref seq found, no bam alignment file generated for lane } . $position );
      return;
    }

  }

  if ( ! $ref_seq ) {
    return;
  }

  my $bsub_command = $self->_alignment_file_bsub_command( {
    position => $position,
    job_dependencies => $arg_refs->{job_dependencies},
    ref_seq => $ref_seq,
    control => $is_control,
    is_paired => $self->is_paired_read(),
    is_spiked_phix => $is_spiked_phix,
  } );

  if ( $self->verbose() ) {
    $self->log( $bsub_command );
  }

  push @{ $job_ids }, $self->submit_bsub_command($bsub_command);

  return;
}

# generate bsub command for generating the alignment files required
sub _alignment_file_bsub_command {
  my ( $self, $arg_refs ) = @_;

  my $position          = $arg_refs->{position};
  my $job_dependencies  = $arg_refs->{job_dependencies};
  my $ref_seq           = $arg_refs->{ref_seq};
  my $is_paired         = $arg_refs->{is_paired};
  my $is_spiked_phix    = $arg_refs->{is_spiked_phix};

  my $mem_size    = $self->general_values_conf()->{bam_creation_memory};
  my $timestamp   = $self->timestamp();
  my $bsub_queue  = $self->lsf_queue;
  my $id_run      = $self->id_run();

  my $job_name  = $self->is_paired_read() ? $self->align_job() . q{_} . $id_run . q{_} . $position . q{_paired_} . $timestamp
                :                           $self->align_job() . q{_} . $id_run . q{_} . $position . q{_} . $timestamp
                ;

  my @command;
  push @command, 'bsub';
  push @command, "-q $bsub_queue";
  push @command, $self->ref_adapter_pre_exec_string();
  push @command, q{-o log/}. $job_name . q{.out};

  push @command, "-J $job_name";
  push @command, '-C ' . $self->dump_size();    # get dump on failure

  push @command, $self->fs_resource_string( {
    resource_string => qq{-R 'select[mem>$mem_size] rusage[mem=$mem_size] span[hosts=1]'},
    counter_slots_per_job => $self->pb_cal_pipeline_conf()->{io_resource_slots},
  } );
  push @command, qq{-M ${mem_size}000};
  push @command,  q{-n } . $self->pb_cal_pipeline_conf()->{io_resource_slots};
  push @command, $job_dependencies || q[];

  push @command,  q["];               # " enclose command in quotes
  push @command, $self->pb_calibration_bin() . q{/} . $self->alignment_script();
  push @command,  q{--aln_parms '-t } . $self->pb_cal_pipeline_conf()->{io_resource_slots} . q{'};

  push @command,  q{--ref } . $ref_seq;

  if( $is_paired ) {
    push @command, q{--read1 } . $self->general_values_conf()->{sanger_read_one};
    push @command, q{--read2 } . $self->general_values_conf()->{sanger_read_two};
  } else {
    push @command, q{--read } . $self->general_values_conf()->{sanger_read_one};
  }

  push @command, q{--bam }.$self->basecall_path().q{/}.$id_run.q{_}.$position.q{.bam};

  push @command, q{--prefix } . $PB_ALIGN_BAM_PREFIX . $id_run.q{_}.$position;

  push @command,  q{--pf_filter};

  push @command, q["];               # " closing quote

  my $bsub_command = join q[ ], @command;
  return $bsub_command;
}



# generate bsub command for recalibrating the lane qseq data
sub _recalibration_bsub_command {
  my ($self, $arg_refs) = @_;
  my $position = $arg_refs->{position};
  my $job_dependencies = $arg_refs->{job_dependencies};

  my $mem_size = $self->mem_score();
  my $timestamp   = $self->timestamp();
  my $bsub_queue  = $self->lsf_queue;
  my $id_run = $self->id_run();

  my $job_name  = $self->score_job() . q{_} . $id_run . q{_} . $position . q{_} . $timestamp;

  my  $cycle_start1 = 1;
  my $cycle_start2;
  if( $self->is_paired_read() ){
    my @r2r = $self->read2_cycle_range();
    $cycle_start2 = $r2r[0];
  }

  my @command;
  push @command, 'bsub';
  push @command, "-q $bsub_queue";
  push @command, q{-o log/}. $job_name . q{.out};

  push @command, "-J $job_name";
  push @command, '-C ' . $self->dump_size();    # get dump on failure

  push @command, $self->fs_resource_string( {
    resource_string => qq{-R 'select[mem>$mem_size] rusage[mem=$mem_size] span[hosts=1]'},
    counter_slots_per_job => $self->pb_cal_pipeline_conf()->{io_resource_slots},
  } );
  push @command, qq{-M ${mem_size}000};
  push @command,  q{-n } . $self->pb_cal_pipeline_conf()->{io_resource_slots};
  push @command, $job_dependencies || q[];

  push @command, q["];               # " enclose command in quotes
  push @command, q{set -o pipefail;};
  push @command, $self->pb_calibration_bin() . q{/} . $self->recalibration_script();
  push @command, q{--u};
  push @command, q{--bam } . $self->basecall_path() . q{/} . $id_run . q{_} . $position . q{.bam};
  #push @command, q{--prefix } . $id_run . q{_} . $position;
  push @command, qq{--run $id_run};
  push @command, q{--runfolder_dir } . $self->runfolder_path();

  if( !$self->is_paired_read() ){
     push @command, qq{--cstart $cycle_start1};
  }else{
     push @command, qq{--cstart1 $cycle_start1};
     push @command, qq{--cstart2 $cycle_start2};
  }

  if ( $arg_refs->{mode} ) {
    push @command, q{--mode } . $arg_refs->{mode};
  }

  if ( $arg_refs->{clmode} ) {
      push @command, q{--clmode } . $arg_refs->{clmode};
  } else {
      push @command, q{--clmode 2}; # assume 2, since most recalibration done with a control will be mode 2 (no per cycle)
  }

  foreach my $key (qw(ct clct ct1 clct1 ct2 clct2)){
      if($arg_refs->{$key}){
          push @command, qq{--$key } . $arg_refs->{$key};
      }
  }

  push @command, qq{| $BAM_MERGER_CMD O=}.$id_run . q{_} . $position . q{.bam};
  push @command, qq{ALIGNED=${PB_ALIGN_BAM_PREFIX}${id_run}_${position}.bam};

  push @command, q["];               # " closing quote

  my $bsub_command = join q[ ], @command;

  return $bsub_command;
}

# generate bsub command for generating the calibration table required
sub _calibration_table_bsub_command {
  my ($self, $arg_refs) = @_;
  my $position = $arg_refs->{position};
  my $read = $arg_refs->{read};
  my $job_dependencies = $arg_refs->{job_dependencies};

  my $control = $arg_refs->{control};

  my $mem_size = $self->mem_calibration();
  my $timestamp   = $self->timestamp();
  my $bsub_queue  = $self->lsf_queue;
  my $id_run = $self->id_run();

  my $job_name  = $self->is_paired_read() ? $self->cal_table_job() . q{_} . $id_run . q{_} . $position . q{_} . $read . q{_} . $timestamp
                :                           $self->cal_table_job() . q{_} . $id_run . q{_} . $position . q{_} . $timestamp
                ;

  my $cycle_start;
  if ($read == 1) {
    $cycle_start = 1;
  } else {
    my @r2r = $self->read2_cycle_range();
    $cycle_start = $r2r[0];
  }

  my @command;
  push @command, 'bsub';
  push @command, "-q $bsub_queue";
  push @command, $self->ref_adapter_pre_exec_string();
  push @command, q{-o log/}. $job_name . q{.out};

  push @command, "-J $job_name";
  push @command, '-C ' . $self->dump_size();    # get dump on failure

  push @command, $self->fs_resource_string( {
    resource_string => qq{-R 'select[mem>$mem_size] rusage[mem=$mem_size] span[hosts=1]'},
    counter_slots_per_job => $self->pb_cal_pipeline_conf()->{io_resource_slots},
  } );
  push @command, qq{-M ${mem_size}000};
  push @command, $job_dependencies || q[];

  push @command, q["];               # " enclose command in quotes
  push @command, $self->pb_calibration_bin() . q{/} . $self->cal_table_script();
  push @command, q{--runfolder_dir } . $self->runfolder_path();
  push @command, q{--intensity_dir }. $self->intensity_path();
  push @command, qq{--run $id_run};
  push @command, qq{--lane $position};
  push @command, q{--t_filter } . $self->t_filter();
  push @command, qq{--read $read};
  push @command, qq{--cstart $cycle_start};
  push @command, q{--prefix } . $id_run . q{_} . $position . q{_} . $read;

  if ( $self->has_pb_cal_mode() ) {
    push @command, q{--mode } . $self->pb_cal_mode();
  } elsif ( $control ) {
    push @command, q{--mode 2};
  } else {};

  if ( $arg_refs->{is_spiked_phix} ) {
    push @command, q{--snp } . $self->pb_cal_pipeline_conf()->{phix_snp_file};
  }

  push @command, qq{--bam ${PB_ALIGN_BAM_PREFIX}${id_run}_${position}.bam};

  push @command, q["];               # " closing quote

  my $bsub_command = join q[ ], @command;
  return $bsub_command;
}

no Moose;

__PACKAGE__->meta->make_immutable;

1;

__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Carp

=item English -no_match_vars

=item Readonly

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2011 GRL, by Guoying Qi (gq1@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
