#############
# $Id: split_bam_by_tag.pm 14173 2011-09-13 09:45:29Z ajb $
# Created By: gq1
# Last Maintained By: $Author: ajb $
# Created On: 2011-08-19
# Last Changed On: $Date: 2011-09-13 10:45:29 +0100 (Tue, 13 Sep 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-30.0/lib/npg_pipeline/analysis/split_bam_by_tag.pm $

package npg_pipeline::analysis::split_bam_by_tag;
use Moose;
use Carp;
use English qw{-no_match_vars};
use Readonly;
use List::Util qw(max);

extends qw{npg_pipeline::analysis::bustard};


Readonly::Scalar our $VERSION  => do { my ($r) = q$LastChangedRevision: 14173 $ =~ /(\d+)/mxs; $r; };
Readonly::Scalar our $LOG2 => log 2;
Readonly::Scalar our $DEFAULT_RESOURCES => 4;

Readonly::Scalar our $SPLIT_BAM_CMD => q{/software/bin/java -jar /software/solexa/bin/aligners/illumina2bam/current/SplitBamByReadGroup.jar TRIM=1 CREATE_MD5_FILE=true VALIDATION_STRINGENCY=SILENT};

sub generate {
  my ($self, $arg_refs) = @_;

  my $id_run = $self->id_run();

  $self->log(qq{Creating Jobs to split bam files by tag for $id_run});

  my @positions = $self->positions( $arg_refs );

  my @job_ids;

  if ( ! $self->is_indexed() ) {
    $self->log( qq{Run $id_run is not multiplex run and no need to split} );
    return;
  }

  my $lane_groups_href = $self->grouping_lanes( @positions );
  my $indexed_lanes = $lane_groups_href->{indexed_lanes};

  my $output_dir = $self->recalibrated_path() . q{/lane};

  foreach my $position ( @{ $indexed_lanes } ) {

    my $lane_output_dir = $output_dir . $position;

    if ( ! -d $lane_output_dir ) {

       $self->log( qq{creating $lane_output_dir} );
       my $rc = `mkdir -p $lane_output_dir`;
       if ( $CHILD_ERROR ) {
         croak qq{could not create $lane_output_dir\n\t$rc};
       }
    }

  }

  my $position = $self->lsb_jobindex();

  $output_dir .= $position;

  # submits a job (in an array) for each multiplex lane
  if ( scalar @{ $indexed_lanes } ) {

    $arg_refs->{array_string} = $self->create_array_string( @{ $indexed_lanes } );
    $arg_refs->{fs_resource} = $self->_max_fs_resource( @{ $indexed_lanes } );
    $arg_refs->{bam} = $self->recalibrated_path().q{/}.$self->id_run().q{_}.$position.q{.bam};
    $arg_refs->{output_prefix} = $output_dir.q{/}.$self->id_run().q{_}.$position;

    my $job_sub = $self->_generate_bsub_command( $arg_refs );
    push @job_ids, $self->submit_bsub_command( $job_sub );

  }

  return @job_ids;
}

###############
# responsible for generating the bsub command to be executed for one lane one fastq file
sub _generate_bsub_command {
  my ($self, $arg_refs) = @_;

  my $required_job_completion = $arg_refs->{required_job_completion};
  my $bam                     = $arg_refs->{bam};
  my $output_prefix           = $arg_refs->{output_prefix};

  my $timestamp = $self->timestamp();

  my $job_name = q{split_bam_by_tag_} . $self->id_run() . q{_} . $timestamp;
  my $replacement = $job_name . q{.%I.%J};

  $job_name .= $arg_refs->{array_string};

  my $outfile = $self->make_log_dir( $self->recalibrated_path() ) . q{/} . $self->bsub_output_file();
  $outfile =~ s/substitute_this/$replacement/xms;

  my $job_command = $SPLIT_BAM_CMD . qq{  I=$bam O=$output_prefix};

  my $job_sub = q{bsub -q } . $self->lsf_queue() . q{ };
  $job_sub .=  ( $self->fs_resource_string( {
                   counter_slots_per_job => $arg_refs->{fs_resource},
               } ) );
  $job_sub .= q{ };
  $job_sub .= qq{$required_job_completion -J $job_name -o $outfile '$job_command'};

  if ($self->verbose()) { $self->log($job_sub); }

  return $job_sub;
}

##################
# work out a LSF fs resource count suitable for splitting the to the max number of plex files 
# for the given lanes. 999 => 39, 168 => 29, 96 => 26, 12 => 14, 8 => 12, 2 => 4, 1 => 4, 0 => 4
sub _max_fs_resource {
  my ($self, @positions) = @_;
  my $fs_resource +=  max map { scalar @{ $self->get_tag_index_list( $_ ) } } @positions;
  $fs_resource = int ( $DEFAULT_RESOURCES*log( $fs_resource < 2 ? 2 : $fs_resource ) / $LOG2 );
  return $fs_resource;
}

sub grouping_lanes {
   my ($self, @positions) = @_;

   my $indexed_lanes     = [];
   my $non_indexed_lanes = [];

   my $lane_groups_hashref = {
     indexed_lanes           => $indexed_lanes,
     non_indexed_lanes       => $non_indexed_lanes,
   };

   foreach my $position ( @positions ) {

     if ( $self->is_multiplexed_lane($position) ) {
        push @{$indexed_lanes}, $position;
      } else {
        push @{$non_indexed_lanes}, $position;
      }
   }

   return $lane_groups_hashref;
}

no Moose;

__PACKAGE__->meta->make_immutable;

1;
__END__

=head1 NAME

npg_pipeline::analysis::split_bam_by_tag

=head1 VERSION

$LastChangedRevision: 14173 $

=head1 SYNOPSIS

  my $oAfgfq = npg_pipeline::archive::file::generation::fastq_by_tag->new({
    run_folder => $sRunFolder,
  });

=head1 DESCRIPTION

Object module which knows how to construct and submits the command line to LSF for splitting bam file by tag.

=head1 SUBROUTINES/METHODS

=head2 generate - generates the bsub jobs and submits them for spliting the bam files by tag, returning an array of job_ids.

  my @job_ids = $oAfgfq->generate({
    required_job_completion} => q{-w (123 && 321)};
    timestamp                => q{20090709-123456};
    solexa_bin               => q{/software/solexa/bin};
    id_run                   => 1234;
    BSUB_ERROR_FILE          => q{substitute_this.err};
    BSUB_OUTPUT_FILE         => q{substitute_this.out};
  });

=head2 grouping_lanes - group lanes


=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Carp

=item English -no_match_vars

=item Readonly

=item Moose

=item List::Util

=item npg_pipeline::analysis::bustard

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2011 GRL, by Guoying Qi (gq1@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
