#############
# $Id: summary.pm 14173 2011-09-13 09:45:29Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2010-01-21
# Last Changed On: $Date: 2011-09-13 10:45:29 +0100 (Tue, 13 Sep 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-30.0/lib/npg_pipeline/archive/file/summary.pm $

package npg_pipeline::archive::file::summary;
use Moose;
use Carp qw{carp croak cluck confess};
use English qw{-no_match_vars};
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 14173 $ =~ /(\d+)/mxs; $r; };
use XML::LibXML;
use XSLT::Cache;
use Perl6::Slurp;

extends qw{npg_pipeline::archive};

Readonly::Scalar our $SCRIPT_NAME  => q{npg_pipeline_munge_summary_multiplex};

# these are related to the xml output from the illumina pipeline, but have been consistent so far
# as such, they are currently remaining hard coded, until they change this
Readonly::Scalar our $SUMMARY_HTML => q{Summary.htm};
Readonly::Scalar our $SUMMARY_XML  => q{Summary.xml};
Readonly::Scalar our $SUMMARY_XSL  => q{Summary.xsl};
Readonly::Scalar our $TIMEOUT      => 100;

Readonly::Scalar our $SUMMARY_TABLE_COLUMNS_TO_MUNGE => {
    LaneResultsSummary => [ qw{averageAlignScorePF errorPF percentUniquelyAlignedPF} ],
    ExpandedLaneSummary => [ qw{errorPF errorRaw infoContentPF infoContentRaw percentUniquelyAlignedPF } ],
    TileResultsByLane => [ qw{averageAlignScorePF errorPF percentUniquelyAlignedPF} ],
  };

=head1 NAME

npg_pipeline::archive::file::summary

=head1 VERSION

$LastChangedRevision: 14173 $

=head1 SYNOPSIS

  my $oSummary = npg_pipeline::archive::file::summary->new({
    recalibrated_path => $sRecalibratedPath,
    id_run => $iIdRun,
    run_folder => $sRunFolder,
  });

=head1 DESCRIPTION

This object is designed to read and munge data from the Gerald output XML file, primarily with the purpose of restructuring
it to match where the data should be for muliplex runs.

From imported run roles, path data can be obtained, as long as one of the 3 has been provided.

=head1 SUBROUTINES/METHODS

=head2 munge

This takes the Summary.htm file, and ensures that the data is in the expected order for the different reads.

  eval {
    $oSummary->munge();
  } or do {
    Your error handling here...
  };

=cut

sub munge {
  my ($self) = @_;

  # read the xml in and create an XML::LibXMl::Document object from it
  my $summary_doc  = $self->recalibrated_path() . qq{/$SUMMARY_XML};
  my $summary_html = $self->recalibrated_path() . qq{/$SUMMARY_HTML};

  my $xml = slurp ( $summary_doc );
  $xml = XML::LibXML->new->parse_string($xml);

  my $lane_results_summary = $xml->getElementsByTagName(q{LaneResultsSummary})->[0];

  my ($read_1, $read_t, $read_2) = $lane_results_summary->getElementsByTagName(q{Read});
  # nothing wil need altering if this is not a paired end multiplex run
  if (!$read_2) {
    return;
  }

  $self->_process_summary_tables($xml, q{LaneResultsSummary});
  $self->_process_summary_tables($xml, q{ExpandedLaneSummary});

  $self->_process_tile_errors_by_lane($xml);
  $self->_process_tile_results_by_lane($xml);

  # store out the xml to file
  $self->_move_to_old_and_write_new({
    file => $summary_doc,
    content => $xml->toString(),
  });

  # recreate the Summary.htm (since people are expecting that)
  my $xslt = XSLT::Cache->new( prefix => $self->recalibrated_path(), timeout => $TIMEOUT );

  $self->_move_to_old_and_write_new({
    file => $summary_html,
    content => $xslt->transform( $summary_doc ),
  });

  return 1;
}

=head2 generate

launch job onto the farm to do this munge

=cut

sub generate {
  my ($self, $arg_refs) = @_;
  my $job_sub = $self->_generate_bsub_command($arg_refs);
  return ( $self->submit_bsub_command( $job_sub ) );
}

##########
# private methods

# move the original file to an old version, and then recreate with the new content
sub _move_to_old_and_write_new {
  my ($self, $arg_ref) = @_;

  my $file     = $arg_ref->{file};
  my $file_old = $file . q{.old};

  my $mv_cmd = qq{mv $file $file_old};
  my $output    = qx[$mv_cmd];

  open my $fh, q{>}, $file or croak qq{Unable to open $file for writing};
  print {$fh} ( $arg_ref->{content} ) or croak q{Unable to print to file};
  close $fh or carp qq{Unable to close filehandle on $file};

  return 1;
}

# the summary tables need alignment data switching from read 2 (the index) to read 3 (the 2nd sequence read)
sub _process_summary_tables {
  my ($self, $xml, $summary_table) = @_;
  my $lane_results_summary = $xml->getElementsByTagName($summary_table)->[0];

  my ($read_1, $read_t, $read_2) = $lane_results_summary->getElementsByTagName(q{Read});

  my $data_store = {};

  my $columns = $SUMMARY_TABLE_COLUMNS_TO_MUNGE->{$summary_table};

  # get data from read t, which should go in read 2
  foreach my $lane ($read_t->getElementsByTagName(q{Lane})) {
    my $lane_number = $lane->getElementsByTagName(q{laneNumber})->[0]->firstChild->nodeValue();

    foreach my $column (@{$columns}) {

      my $column_val = $lane->getElementsByTagName($column)->[0];

      if ($column_val) {
        if (!$data_store->{$lane_number}) {
          $data_store->{$lane_number} = {};
        }
        $data_store->{$lane_number}->{$column} = $column_val;
      }
    }
  }

  # move node to the read_2 node
  foreach my $lane ($read_2->getElementsByTagName(q{Lane})) {
    my $lane_number = $lane->getElementsByTagName(q{laneNumber})->[0]->firstChild->nodeValue();

    if ($data_store->{$lane_number}) {
      foreach my $column (@{$columns}) {
        $lane->appendChild( $data_store->{$lane_number}->{$column} );
      }
    }
  }

  return 1;
}

# _process_tile_errors_by_lane - read 2 needs renaming read 3
sub _process_tile_errors_by_lane {
  my ($self, $xml) = @_;
  my $table = $xml->getElementsByTagName(q{TileErrorsByLane})->[0];
  my @lanes = $table->getElementsByTagName(q{Lane});

  foreach my $lane ( @lanes ) {

    my @reads = $lane->getElementsByTagName(q{Read});

    foreach my $read ( @reads ) {
      my $read_number_node = $read->getElementsByTagName(q{readNumber})->[0];
      my $read_number = $read_number_node->firstChild()->nodeValue();
      if ( $read_number == 2 ) {
        $read_number_node->removeChildNodes();
        $read_number_node->appendTextNode( $self->general_values_conf()->{illumina_read_two} );
      }
    }

  }
  return 1;
}

# TileResultsByLane - tile by tile info - the alignment data needs moving from read 2 (the index) to read 3 (the 2nd sequence read)
sub _process_tile_results_by_lane {
  my ($self, $xml) = @_;
  my $table = $xml->getElementsByTagName(q{TileResultsByLane})->[0];
  my @lanes = $table->getElementsByTagName(q{Lane});

  my $data_store = {};

  my $columns = $SUMMARY_TABLE_COLUMNS_TO_MUNGE->{TileResultsByLane};

  foreach my $lane ( @lanes ) {

    my $lane_number = $lane->getElementsByTagName(q{laneNumber})->[0]->firstChild()->nodeValue();

    my ($read1, $readt, $read2) = $lane->getElementsByTagName(q{Read});

    my $has_alignment_data = 0; # get out of loops quickly if no data to munge about!

    foreach my $tile ( $readt->getElementsByTagName(q{Tile}) ) {

      my $read_number = $readt->getElementsByTagName(q{readNumber})->[0]->firstChild()->nodeValue();

      my $tile_number = $tile->getElementsByTagName(q{tileNumber})->[0]->firstChild()->nodeValue();

      foreach my $column (@{$columns}) {

        my $column_val = $tile->getElementsByTagName($column)->[0];

        if ($column_val) {

          $has_alignment_data++;

          if (!$data_store->{$lane_number}) {
            $data_store->{$lane_number} = {};
          }
          $data_store->{$lane_number}->{$tile_number}->{$column} = $column_val;
        }

        if (!$has_alignment_data) {
          last;
        }

      }

      if (!$has_alignment_data) {
        last;
      }

    }

    if (!$has_alignment_data) {
      next;
    }

    foreach my $tile ( $read2->getElementsByTagName(q{Tile}) ) {

      my $read_number = $read2->getElementsByTagName(q{readNumber})->[0]->firstChild()->nodeValue();

      my $tile_number = $tile->getElementsByTagName(q{tileNumber})->[0]->firstChild()->nodeValue();

      if ($data_store->{$lane_number}->{$tile_number}) {

        foreach my $column (@{$columns}) {
          $tile->appendChild( $data_store->{$lane_number}->{$tile_number}->{$column} );
        }
      }
    }

  }

  return 1;
}

# for generating the bsub job command which will relaunch this

sub _generate_bsub_command {
  my ($self, $arg_refs) = @_;
  my $required_job_completion = $arg_refs->{required_job_completion};

  my $script_bin = $self->internal_script_bin();
  my $id_run = $self->id_run();

  my $job_name = $SCRIPT_NAME . q{_} . $id_run . q{_} . $self->timestamp();

  my $outfile  = $self->make_log_dir( $self->recalibrated_path() ) . q{/} . $job_name . q{.out};

  my $options = $self->attributes_as_command_options({
    excluded_attributes => [qw(
        time timestamp lsf_queue script_bin run batch default_usr_bin
        summary_link max_lane owning_group mail_from control_lane_position default_control_lane internal_script_bin
        bsub_error_file bsub_output_file subpath log_file log_file_name log_file_path script_name array_cpu_limit
      )],
  });

  my $job_sub = q{bsub -q } . $self->lsf_queue() . q{ };
  $job_sub .=  ( $self->fs_resource_string( {
    counter_slots_per_job => 1,
  } ) ) . q{ };
  $job_sub .= qq{$required_job_completion -J $job_name -o $outfile '$script_bin/$SCRIPT_NAME $options'};

  if ( $self->verbose() ) {
    $self->log($job_sub);
  }

  return $job_sub;
}

no Moose;
__PACKAGE__->meta->make_immutable;
1;
__END__


=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Carp

=item English -no_match_vars

=item Readonly

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
