#############
# $Id: UpdateLane.pm 14173 2011-09-13 09:45:29Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2011-09-05
# Last Changed On: $Date: 2011-09-13 10:45:29 +0100 (Tue, 13 Sep 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-30.0/lib/npg_pipeline/run/status/UpdateLane.pm $

package npg_pipeline::run::status::UpdateLane;
use Moose;
use Carp;
use English qw{-no_match_vars};
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 14173 $ =~ /(\d+)/mxs; $r; };
use DateTime;

extends qw{npg_pipeline::run::status};
with qw{
  npg_pipeline::roles::business::databaseConnection
  npg_pipeline::roles::business::pre_exec_strings
};

Readonly::Scalar our $SCRIPT_NAME => q{npg_pipeline_update_lane_status};

=head1 NAME

npg_pipeline::run::status::UpdateLane

=head1 VERSION

$LastChangedRevision: 14173 $

=head1 SYNOPSIS

  use npg_pipeline::archive::file::generation;

  my $oULS = npg_pipeline::run::status::UpdateLane->new({
    id_run => 1234,
    position => 1,
    lane => 1,
    status => 'analysis complete',
  });

  position or lane can be provided - position is used in preference

=head1 DESCRIPTION

This class updates the run lane status to analysis complete through the DBIx::Class interface to the npg_tracking database.

=head1 SUBROUTINES/METHODS

=head2 update

Connects to the database and updates the status, via the update_run_lane_status on the RunLaneStatus object.

Takes no arguments, as the information should already be in the object on construction

=cut

has q{status} => (
  isa => q{Str},
  is  => q{ro},
  required => 1,
);

has q{position} => (
  isa => q{Int},
  is  => q{rw},
);

sub update {
  my ( $self ) = @_;

  my $position = $self->position() || $self->lane();

  if ( ! $position ) {
    croak q{No lane/position provided};
  }

  eval {
    my $schema = $self->dbix_tracking_schema();
    my $rl = $schema->resultset( q{RunLane} )->search( {
      id_run => $self->id_run(),
      position => $position,
    } )->first();
    my $crls = $rl->current_run_lane_status();
    if ( $crls ) {
      $crls->update_run_lane_status( {
        id_run_lane => $rl->id_run_lane(),
        username => q{pipeline},
        description => $self->status(),
      } );
    } else {
      my $user = $schema->resultset( q{User} )->search( {username => q{pipeline}, } )->first();
      my $rlsd = $schema->resultset( q{RunLaneStatusDict} )->search( {description => $self->status(), } )->first();
      $schema->resultset( q{RunLaneStatus} )->create( {
        id_run_lane => $rl->id_run_lane(),
        id_user => $user->id_user(),
        id_run_lane_status_dict => $rlsd->id_run_lane_status_dict(),
        date => DateTime->now(),
      } );
    }
    $self->log( q{Updated Run: } . $self->id_run() . q{ Lane: } . $position . q{ to } . $self->status() );
    1;
  } or do {
    croak $EVAL_ERROR;
  };

  return 1;
}

=head2 launch_bsub_jobs

generates the bsub jobs and submits them for updating lane status. Returns an array of job ids for each of the submitted lanes

=cut

sub launch_bsub_jobs {
  my ( $self, $arg_refs ) = @_;


  my @job_ids;
  my @positions = $self->positions( $arg_refs );

  foreach my $lane ( @positions ) {
    $arg_refs->{lane} = $lane;
    my $bsub_command = $self->_generate_bsub_command( $arg_refs );
    push @job_ids, $self->submit_bsub_command( $bsub_command );
  }

  return @job_ids;

}

sub _generate_bsub_command {
  my ( $self, $arg_refs ) = @_;

  my $required_job_completion = $arg_refs->{required_job_completion};
  my $timestamp   = $self->timestamp();
  my $script_bin  = $self->internal_script_bin();
  my $id_run      = $self->id_run();

  my $outfile     = $self->make_log_dir( $self->archive_path() ) . q{/} . $self->bsub_output_file();
  my $job_name    = q{update_run_lane_status_} . $id_run . q{_} . $arg_refs->{lane} . q{_} . $timestamp;
  my $replacement = $job_name . q{.%I.%J};
  $outfile        =~ s/substitute_this/$replacement/xms;

  my $options = $self->attributes_as_command_options({
    excluded_attributes => [ qw(
      log_file log_file_name log_file_path subpath casava_pipeline illumina_pipeline timestamp
      bustard_home gerald_config_prefix runtype script_name time bsub_output_file lane lanes
      ) ],
      quotes => 1,
  });

  my $job_sub = q{bsub -q } . $self->lsf_queue() . q{ };

  $job_sub .= $self->tracking_dbix_connection_pre_exec_string() . q{ };

  $job_sub .= qq{$required_job_completion -J $job_name -o $outfile '$script_bin/} . $SCRIPT_NAME . q{ } . $options;

  $job_sub .= q{ --position "} . $arg_refs->{lane} . q{"'};

  if ( $self->verbose() ) { $self->log( $job_sub ); }
  return $job_sub;
}


no Moose;
__PACKAGE__->meta->make_immutable;
1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Carp

=item English -no_match_vars

=item Readonly

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2011 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
